@@ -95,11 +95,11 @@ static adv_error mng_read_ihdr(adv_mng* mng, adv_fz* f, const unsigned char* ihd
 
 	mng->dat_width = be_uint32_read(ihdr + 0);
 	mng->dat_height = be_uint32_read(ihdr + 4);
-	if (mng->dat_x + mng->frame_width > mng->dat_width) {
+	if (mng->dat_x < 0 || mng->dat_x + mng->frame_width > mng->dat_width) {
 		error_set("Frame not complete");
 		goto err;
 	}
-	if (mng->dat_y + mng->frame_height > mng->dat_height) {
+	if (mng->dat_y < 0 || mng->dat_y + mng->frame_height > mng->dat_height) {
 		error_set("Frame not complete");
 		goto err;
 	}
@@ -340,6 +340,15 @@ static adv_error mng_read_defi(adv_mng* mng, unsigned char* defi, unsigned defi_
 		mng->dat_y = 0;
 	}
 
+	if (mng->dat_x < 0 || mng->dat_x >= (int)mng->frame_width) {
+		error_set("Invalid move");
+		return -1;
+	}
+	if (mng->dat_y < 0 || mng->dat_y >= (int)mng->frame_height) {
+		error_set("Invalid move");
+		return -1;
+	}
+
 	return 0;
 }
 
@@ -347,6 +356,8 @@ static adv_error mng_read_move(adv_mng* mng, adv_fz* f, unsigned char* move, uns
 {
 	unsigned id;
 
+	(void)f;
+
 	if (move_size != 13) {
 		error_unsupported_set("Unsupported MOVE size in MOVE chunk");
 		return -1;
@@ -375,10 +386,19 @@ static adv_error mng_read_move(adv_mng* mng, adv_fz* f, unsigned char* move, uns
 		return -1;
 	}
 
+	if (mng->dat_x < 0 || mng->dat_x + mng->frame_width > mng->dat_width) {
+		error_set("Invalid move");
+		return -1;
+	}
+	if (mng->dat_y < 0 || mng->dat_y + mng->frame_height > mng->dat_height) {
+		error_set("Invalid move");
+		return -1;
+	}
+
 	return 0;
 }
 
-static void mng_delta_replacement(adv_mng* mng, unsigned pos_x, unsigned pos_y, unsigned width, unsigned height)
+static int mng_delta_replacement(adv_mng* mng, unsigned dlt_size, unsigned pos_x, unsigned pos_y, unsigned width, unsigned height)
 {
 	unsigned i;
 	unsigned bytes_per_run = width * mng->pixel;
@@ -386,14 +406,19 @@ static void mng_delta_replacement(adv_mng* mng, unsigned pos_x, unsigned pos_y,
 	unsigned char* p0 = mng->dat_ptr + pos_y * mng->dat_line + pos_x * mng->pixel + 1;
 	unsigned char* p1 = mng->dlt_ptr + 1;
 
+	if (dlt_size != delta_bytes_per_scanline * height)
+		return -1;
+
 	for(i=0;i<height;++i) {
 		memcpy(p0, p1, bytes_per_run);
 		p0 += mng->dat_line;
 		p1 += delta_bytes_per_scanline;
 	}
+
+	return 0;
 }
 
-static void mng_delta_addition(adv_mng* mng, unsigned pos_x, unsigned pos_y, unsigned width, unsigned height)
+static int mng_delta_addition(adv_mng* mng, unsigned dlt_size, unsigned pos_x, unsigned pos_y, unsigned width, unsigned height)
 {
 	unsigned i, j;
 	unsigned bytes_per_run = width * mng->pixel;
@@ -401,6 +426,9 @@ static void mng_delta_addition(adv_mng* mng, unsigned pos_x, unsigned pos_y, uns
 	unsigned char* p0 = mng->dat_ptr + pos_y * mng->dat_line + pos_x * mng->pixel + 1;
 	unsigned char* p1 = mng->dlt_ptr + 1;
 
+	if (dlt_size != delta_bytes_per_scanline * height)
+		return -1;
+
 	for(i=0;i<height;++i) {
 		for(j=0;j<bytes_per_run;++j) {
 			*p0++ += *p1++;
@@ -408,6 +436,8 @@ static void mng_delta_addition(adv_mng* mng, unsigned pos_x, unsigned pos_y, uns
 		p0 += mng->dat_line - bytes_per_run;
 		p1 += delta_bytes_per_scanline - bytes_per_run;
 	}
+
+	return 0;
 }
 
 static adv_error mng_read_delta(adv_mng* mng, adv_fz* f, unsigned char* dhdr, unsigned dhdr_size)
@@ -538,16 +568,22 @@ static adv_error mng_read_delta(adv_mng* mng, adv_fz* f, unsigned char* dhdr, un
 			goto err_data;
 		}
 
-		dlt_size = mng->dat_size;
+		dlt_size = mng->dlt_size;
 		if (uncompress(mng->dlt_ptr, &dlt_size, data, size) != Z_OK) {
 			error_set("Corrupt compressed data in IDAT chunk");
 			goto err_data;
 		}
 
 		if (ope == 0 || ope == 4) {
-			mng_delta_replacement(mng, pos_x, pos_y, width, height);
+			if (mng_delta_replacement(mng, dlt_size, pos_x, pos_y, width, height) != 0) {
+				error_unsupported_set("Invalid delta");
+				goto err_data;
+			}
 		} else if (ope == 1) {
-			mng_delta_addition(mng, pos_x, pos_y, width, height);
+			if (mng_delta_addition(mng, dlt_size, pos_x, pos_y, width, height) != 0) {
+				error_unsupported_set("Invalid delta");
+				goto err_data;
+			}
 		} else {
 			error_unsupported_set("Unsupported delta operation");
 			goto err_data;
